# RLE-DTW Implementation

## How to build

- `mkdir build.build`
- `cd build.build`
- `cmake ../src`
- `make`

## How to use

- Build (see `How to build` above)

- Sequence generating:

    `./dtwGenerator -h` to see the available options
  
    Example run:  
        ```./dtwGenerator -l 20 -p 5```  
    Generates two sequences, each containing 20 runs, whose lengths follow a Poisson(5) distribution

    Example run 2:  
        ```./dtwGenerator -n 10 -v 0 1 -l 100 -u 1 17```  
    Generates 10 binary sequences, each containing 100 runs, whose lengths follow a Discrete-Uniform[1, 17] distribution

- Benchmarking:
    
    `./dtwBenchmark -h` to see the available options

    Example run:  
        ```./dtwBenchmark -l 5 -u -n 1 -c 0.4 -f test.txt```  
    Runs the benchmark on the first 5 sequences (in UCR format) in the file `test.txt`, using only the naive, BDTW and O(kappa) algorithms (BDTW and O(kappa) are toggled on by default). Compresses each sequence to 40% of its size before running the benchmarks.

    Example run 2:  
        ```./dtwGenerator -n 5 -l 50 -u 1 20 | ./dtwBenchmark -a 1 -i```
    Generates 5 sequences of integers, each containing 50 runs with lengths distributed uniformly within [1, 20], and then runs a benchmark on them with BDTW, O(kappa) algorithm and AWarp.

## Structure of the project

The project consists of two main parts:

- Sequence generator (dtwGenerator), which allows to generate sequences of runs whose lengths follow some given distribution.

- Benchmark (dtwBenchmark), which takes an arbitrary number of sequences as input and computes the (approximation of the) DTW distance between all pairs of sequences using the selected algorithms and computes statistics which allow for a comparison between the chosen algorithms.

Apart from the source files for the two executables, the project contains:
- implementation of the O(kappa) algorithm (in folder ../lin\_kapp\_impl),
- implementations of some other algorithms (dtwAlgorithms.\*, awarp.\*, sparsedtw.\*),
- some files contain a legacy algorithm not used in the project (dtwAlgorithms.\*, intersectionPoints.\*),
- a DP compression algorithm (compression.\*) which increases the sparsity ratio by grouping neighbouring values and asigning a single value to them,
- a utilty file (utility.\*) which contains miscellaneous utility functions used throughout the project
