#include "compression.h"

    /*
    //auto x = std::vector<double>{1.0, 0.0, 3.0, 1.0, 5.0, 0.0, 0.0, 3.0, 7.0, 5.0};
	//auto x = std::vector<double>{-0.71052, -1.1833, -1.3724, -1.5931, -1.467, -1.3724, -1.0888, 0.045967, 0.92853, 1.0861, 1.2753, 0.96005, 0.61333, 0.014447, -0.64748, -0.26923, -0.20619, 0.61333, 1.3698, 1.4644, 1.0546, 0.58181, 0.17205, -0.26923};

	auto x = std::vector<double>{-0.50952, -1.2396, -1.2396, -1.6046, -1.6046, -1.1483, -0.78329, -1.2396, -0.14449, 0.95059, 1.3156, 1.4981, 1.2244, 0.85933, 0.49431, 0.76808, 0.49431, 0.49431, 0.12928, -0.327, -0.69203, 1.0418, 0.85933, 0.40305};

	//z0 = [-1.0570666666666666, 1.169604, 0.476058, -0.509515, 0.7680600000000001]
	//z1 = [-1.17113875, -0.14448999999999934, 1.169604, 0.1944657142857143, 0.7680600000000001]
	
    //int n = 1600;
	//auto xir = akt::randomSequence(std::uniform_int_distribution<>(0, 10), std::uniform_int_distribution<>(1, 2), n);
    //std::vector<double> x;
    //for (const auto& r : xir)
    //  x.push_back(r.val);
	
	auto c = akt::compress(x, 5);

	for (const auto& v : c) {
	    for (auto d : v)
            std::cout << d << ", ";
        std::cout << '\n';
    */

namespace {
    /**
    * Some simple helper classes to compute intertia and centroids in constant time.
    */

    class CCMatrix 
    {
    public:    
        CCMatrix(std::vector<double>* sum, std::vector<double>* ssq)
            : sum(sum), ssq(ssq)
        { }

        double get(int i, int j) const
        {
            if (i >= j) return 0.0;
            
            double t = j - i + 1;
            if (i > 0) {
                double s = (*sum)[j] - (*sum)[i-1];
                double mu =  s / t;
                return t * mu * mu - ((mu + mu) * s) + ((*ssq)[j] - (*ssq)[i-1]);
            } else {
                double mu =  (*sum)[j] / t;
                return t * mu * mu - ((mu + mu) * (*sum)[j]) + (*ssq)[j];
            }
        }

        double centroid(int i, int j) const
        {
            if (j == 0) return (*sum)[0];
            double t = (double)(j-i+1);
            double s = (*sum)[j];
            if (i > 0)
                s -= (*sum)[i - 1];
            return s / t;
        }

    private:
        std::vector<double>* sum;
        std::vector<double>* ssq;
    };

    class CiMatrix 
    {
    public:
        CiMatrix(std::vector<std::vector<double>>* d, CCMatrix* M)
            : D(d), CC(M)
        { }
        
        double get(int i, int m, int j) const
        {
            if (j == 0) return CC->get(0, m);
            if (j > m) return (*D)[i-1][m];
            return (*D)[i-1][j - 1] + CC->get(j,m);
        }
        
    private:
        std::vector<std::vector<double>>* D;
        CCMatrix* CC;
    };
    
    /**
    * SMAWK algorithm to compute the leftmost column index of a minimum for each row of a totally
    * monotone (m x n)-matrix. Running time is O(m+n).
    *
    * Reference:
    * A. Aggarwal, M. Klawe, S. Moran, P. Shor and R. Wilber,
    * Geometric applications of a matrix searching algorithm,
    * Algorithmica, 2:195-208, 1987.
    */
    std::vector<int> smawk(const CiMatrix& M, int i, const std::vector<int>& rows, const std::vector<int>& cols) 
    {

        int n = rows.size();

        //end of recursion
        if (n == 0) return std::vector<int>();

        //reduce phase: filter out dead columns
        std::vector<int> stack;
        for (int c : cols) {
            while ((!stack.empty()) && (M.get(i, rows[stack.size() - 1], stack[stack.size() - 1]) > M.get(i, rows[stack.size() - 1], c)))
                stack.pop_back();
            
            if (stack.size() != n)
                stack.push_back(c);
        }
        
        //recursively search indices for every row with odd index
        std::vector<int> temprows;
        for (int j = 1; j < n; j += 2)
            temprows.push_back(rows[j]);
        
        std::vector<int> sub = smawk(M, i, temprows, stack);

        //fill in indices for each row with even index
        std::vector<int> result;
        
        int c = 0;
        for (int j = 0; j < n; j += 2) {
            int r = rows[j];
            int cc = c;
            if (j == n - 1) {
                cc = stack.size() - 1;
                int l = c;
                for (int x = c + 1; x <= cc; x++) {
                    if (M.get(i, r, stack[x]) < M.get(i, r, stack[l]))
                        l = x;
                }
                result.push_back(stack[l]);
            } else {
                int target = sub[j / 2];
                while (stack[cc] != target)
                    cc++;
                int l = c;
                for (int x = c + 1; x <= cc; x++) {
                    if (M.get(i, r, stack[x]) < M.get(i, r, stack[l]))
                        l = x;
                }
                result.push_back(stack[l]);
                result.push_back(target);
            }
            c = cc;
        }
        
        return result;
    }
} // end anonymous namespace

namespace akt 
{
    /**
    * Implementation of an O(nk)-time dynamic program based on SMAWK algorithm.
    * Note that the input is not sorted before.
    * Outputs a compression (h-means clustering) of x for each value h = 1...k.
    *
    * Reference:
    * Xiaolin Wu,
    * Optimal quantization by matrix searching,
    * Journal of Algorithms, 12(4):663-673, 1991.
    */
	template <typename Type>
    std::vector<std::vector<std::pair<int, double>>> compress(const std::vector<Type>& x, int k)
    {
        int n = x.size();

        if (n < k)
            k = n;

        //compute prefix sum arrays
        auto sum = std::vector<double>(n);
        auto sumsq = std::vector<double>(n);

        sum[0] = x[0];
        sumsq[0] = static_cast<double>(x[0]) * x[0];
        
        for (int i = 1; i < n; i++) {
            sum[i] = sum[i - 1] + x[i];
            sumsq[i] = sumsq[i - 1] + static_cast<double>(x[i]) * x[i];
        }

        //create data structures
        auto CC = CCMatrix(&sum, &sumsq);
        
        auto D = std::vector<std::vector<double>>(k, std::vector<double>(n));
        auto T = std::vector<std::vector<int>>(k, std::vector<int>(n));

        auto M = CiMatrix(&D, &CC);
        
            //fill DP tables
        for (int m = 0; m < n; m++) {
            D[0][m] = CC.get(0, m);
            T[0][m] = 0;
        }	
        for (int i = 1; i < k; i++) {
            for (int m = 0; m <= i; m++) {
                D[i][m] = 0.0;
                T[i][m] = m;
            }

            std::vector<int> rows;
            std::vector<int> cols;
            for (int r = i + 1; r < n; r++) rows.push_back(r);
            for (int c = 0; c < n; c++) cols.push_back(c);

            std::vector<int> minima = smawk(M, i, rows, cols);

            int m = i + 1;
            for (int c : minima) {
                T[i][m] = c;
                D[i][m] = M.get(i, m, c);
                m++;
            }
        }
        /*for (int i = 0; i < k; ++i) {
            for (int j = 0; j < n; ++j) {
                std::cout << i << ' ' << j << ' ' << T[i][j] << std::endl;
                if (T[i][j] > j)
                    std::cerr << "wut: " << i << ' ' << j << ' ' << T[i][j] << std::endl;
            }
        }*/
        
        //backtrack to create compressed series
		std::vector<std::vector<std::pair<int, double>>> result;
        for (int i = 0; i < k; i++) {
            int u = n - 1;
            int l = n;
            int j = i;
            std::vector<std::pair<int, double>> compressed;
            while (l > 0) {
                l = T[j][u];
                //if (i == (k-1) && (u < l))
                //    std::cerr << "wutlol: " << j << ' ' << u << ' ' << T[j][u] << ' ' << l << std::endl;
                //    std::cout << "wutlol: " << j << ' ' << u << ' ' << l << std::endl;
                compressed.emplace_back(u - l + 1, CC.centroid(l, u));
                u = l - 1;
                j--;
            }
            result.emplace_back(compressed.crbegin(), compressed.crend());
        }
        
        return result;
    }   

	// Compress into x into k runs (or min(k, x.size()), to be precise) using the algorithm implemented in compress()
	template <typename T>
	std::vector<Run<double>> compressToRuns(const std::vector<T>& x, int k)
	{
		k = std::max(0, std::min(static_cast<int>(x.size()), k));

		auto somewhatCompressed = compress(x, k);
		int blockEnd = -1;
		std::vector<Run<double>> res;
		for (auto p : somewhatCompressed[k - 1]) {
			blockEnd += p.first;
			res.emplace_back(p.second, blockEnd, p.first);
		}

		return res;
	}

	// Needed for the code to link (a bit of a hack but we only use ints, int64_t's and doubles, anyway
	template std::vector<std::vector<std::pair<int, double>>> compress<int>(const std::vector<int>& x, int k);
    template std::vector<std::vector<std::pair<int, double>>> compress<int64_t>(const std::vector<int64_t>& x, int k);
    template std::vector<std::vector<std::pair<int, double>>> compress<double>(const std::vector<double>& x, int k);
	
	template std::vector<Run<double>> compressToRuns<int>(const std::vector<int>& x, int k);
	template std::vector<Run<double>> compressToRuns<int64_t>(const std::vector<int64_t>& x, int k);
	template std::vector<Run<double>> compressToRuns<double>(const std::vector<double>& x, int k);

} // end namespace akt
