#pragma once

#include <vector>

#include "intersectionPoints.h"
#include "utility.h"

namespace akt {
    
	// Computes the square of the DTW-distance between x and y using the standard DP algorithm
    template<typename T>
	T dtwDistance_basicDP(const std::vector<T>& x, const std::vector<T>& y);

	// Uses the O(kl*log(kl) + \kappa) algorithm to compute the DTW distance
    template<typename T>
	T dtwDistance_RLEDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y);

	// Computes the dtwDistance using the RLEDTW algorithm given the intersection points
    template<typename T>
	T dtwDistance_RLEDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, Intersections<T> ips);

	// Implements the BlockDTW algorithm (which considers only the corners of boundaries in its calculations
	// If max is true, it provides an upper bound on the DTW distance; otherwise it provides a lower bound on the DTW distance
    template<typename T>
	T dtwDistanceApp_blockDTW(const std::vector<Run<T>>& x, const std::vector<Run<T>>& y, bool max);

}	// end namespace akt
