#include <algorithm>
#include <numeric>
#include <vector>

namespace akt {
    
    // Struct for storing observations of a sequence (i.e. the non-zero elements of the sequence)
    template <typename T>
    struct SparseObservation {
        SparseObservation(T val, int pos) 
            : val(val), pos(pos) { }
        // The value of the observation
        T val;
        // The position of the observation in the full, zero-containing sequence (i.e. for the 5 in "30050100202" pos would be 3)
        int pos;
    };

    // Converts a "flat" signal to the SparseObservation format
    template <typename T>
    std::vector<SparseObservation<T>> toSparseObservation(const std::vector<T>& x);

    // Runs the SparseDTW (SDTW) algorithm on the given sequence and returns the computed DTW distance
    // Note: if one of the original (flat) observations ended on a row of 0's but the other one didn't the following function may give a slightly innaccurate result
    template <typename T>
    T sparseDtw(const std::vector<SparseObservation<T>>& x, const std::vector<SparseObservation<T>>& y);

} // end namespace akt
