
#pragma once

#include<iostream>

template<class T = double>
struct Cell
{
  const int_fast32_t x,y;
  const T cost;

  inline int_fast32_t x_offset() const { return x - y; }

  Cell(const int_fast32_t _x, const int_fast32_t _y, const T _cost):
    x(_x), y(_y), cost(_cost)
  {}

  inline T horizontal_cost(const int_fast32_t _x, const T block_cost) const { return cost + (_x - x) * block_cost; }
  inline T vertical_cost(const int_fast32_t _y, const T block_cost) const { return cost + (_y - y) * block_cost; }

  friend std::ostream& operator<<(std::ostream& os, const Cell& c) {  return os<<"[("<<c.x<<","<<c.y<<"), "<<c.cost<<"]"; }
};

template<class T = double>
using CellIter = typename std::list<Cell<T>>::const_iterator;
template<class T = double>
using RevCellIter = typename std::reverse_iterator<typename std::list<Cell<T>>::const_iterator>;

struct RawBlock
{
  const int_fast32_t bottom, left, top, right;
  inline int_fast32_t x_offset() const { return right - top; }

  RawBlock(const int_fast32_t _bottom, const int_fast32_t _left, const int_fast32_t _top, const int_fast32_t _right):
    bottom(_bottom), left(_left), top(_top), right(_right)
  {}
};

template<class T = double>
struct Block: public RawBlock
{
  const T cost;

  Block(const int_fast32_t _bottom, const int_fast32_t _left, const int_fast32_t _top, const int_fast32_t _right, const T _cost):
    RawBlock(_bottom, _left, _top, _right),
    cost(_cost)
  {}
};

template<class T = double>
using BlockVec = std::vector<Block<T>>;


