
#pragma once

#include "utils.hpp"

namespace kn{


  template<class T = double>
  using RowTable = T*;
  template<class T = double>
  using ColTable = std::vector<std::vector<T>>;


  uint_fast32_t runs_x, runs_y;
  uint_fast32_t nx, ny;


  template<class T = double>
  void treat_block_row(RowTable<T>& row_table, ColTable<T>& col_table, const uint_fast32_t j, const uint_fast32_t top_index, const BlockInfo<T>& blocks)
  {
    uint_fast32_t right_index = 0;
    T last_cost = col_table[0].back();
    const T _cost = blocks.cost(0, j);
    for(uint_fast32_t y = 0; y != blocks.height(j); ++y)
      col_table[0].push_back(last_cost += _cost);

    DEBUG(std::cout << "\n---treating block-row["<<j<<"]---\n");
    for(uint_fast32_t i = 0; i != runs_x; ++i){
      const T cost = blocks.cost(i, j);
      const T width_cost = cost * blocks.width(i);
      const std::vector<T>& left_col = col_table[i];
      right_index += blocks.width(i);

      std::vector<T>& right_col = col_table[i + 1];
      last_cost = right_col.back();
      for(uint_fast32_t y = 1; y != blocks.height(j); ++y){
        last_cost += cost; // <-- vertical
        //std::cout << "with vert-cost = "<<last_cost<<'\n';
        last_cost = std::min(last_cost, width_cost + left_col[right_col.size()]); // <-- horizontal
        //std::cout << "with hori-cost = "<<last_cost<<"\t (from col["<<i<<"]["<<right_col.size()<<"] = "<<left_col[right_col.size()]<<")\n";
        last_cost = std::min(last_cost, ((y < blocks.width(i)) ? row_table[right_index - y] + cost * y : left_col[right_col.size() - blocks.width(i)] + width_cost));
        //std::cout << "with diag-cost = "<<last_cost<<'\n';
        DEBUG(std::cout << "set (col)-cost["<<right_index<<","<<right_col.size()<<"] = "<<last_cost<<'\n');
        right_col.push_back(last_cost);
      }
    }

    RowTable<T> next_row = new T[nx + 1];
    last_cost = next_row[0] = row_table[0] + blocks.height(j) * blocks.cost(0,j);
    uint_fast32_t index = 1;
    for(uint_fast32_t i = 0; i != runs_x; ++i){
      const T cost = blocks.cost(i, j);
      const T height_cost = cost * blocks.height(j);
      const std::vector<T>& left_col = col_table[i];
      for(uint_fast32_t x = 1; x != blocks.width(i); ++x){
        last_cost += cost; // <-- horizontal
        //std::cout << "with hori-cost = "<<last_cost<<'\n';
        last_cost = std::min(last_cost, height_cost + row_table[index]); // <-- vertical
        //std::cout << "with vert-cost = "<<last_cost<<'\n';
        last_cost = std::min(last_cost, (x < blocks.height(j)) ? left_col[top_index - x] + cost * x : row_table[index - blocks.height(j)] + height_cost);
        //std::cout << "with diag-cost = "<<last_cost<<'\n';
        DEBUG(std::cout << "set (row)-cost["<<index<<"] = "<<last_cost<<'\n');
        next_row[index++] = last_cost;
      }
      last_cost += cost; // <-- horizontal
      last_cost = std::min(last_cost, col_table[i + 1].back() + cost); // <-- vertical
      last_cost = std::min(last_cost, (blocks.width(i) < blocks.height(j)) ?
          left_col[top_index - blocks.width(i)] + blocks.width(i) * cost :
          row_table[index - blocks.height(j)] + height_cost);
      DEBUG(std::cout << "set (row)-cost["<<index<<"] = "<<last_cost<<'\n');
      next_row[index++] = last_cost;
      col_table[i + 1].push_back(last_cost);
    }
    delete[] row_table;
    row_table = next_row;
  }
  

  // simple k*n dynamic programming
  template<class T = double>
  T compute_dtw(const RLEncoding<T>& _x_seq, const RLEncoding<T>& _y_seq)
  {
    runs_x = _x_seq.size();
    runs_y = _y_seq.size();

    const BlockInfo<T> blocks(_x_seq, _y_seq);

    nx = 0; ny = 0;
    for(uint_fast32_t j = 0; j < runs_y; ++j) ny += blocks.height(j);
    for(uint_fast32_t i = 0; i < runs_x; ++i) nx += blocks.width(i);

    RowTable<T> row_table = new T[nx + 1];
    ColTable<T> col_table;
    col_table.resize(runs_x + 1);

    row_table[0] = 0;
    col_table[0].reserve(ny + 1);
    col_table[0].push_back(0);

    uint_fast32_t index = 0;
    T last_cost = 0;
    for(uint_fast32_t i = 0; i != runs_x; ++i){
      const T cost = blocks.cost(i, 0);
      for(uint_fast32_t x = 0; x != blocks.width(i); ++x)
        row_table[++index] = (last_cost += cost);
      col_table[i + 1].reserve(ny + 1);
      col_table[i + 1].push_back(row_table[index]);
    }

    uint_fast32_t current_y = 0;
    for(uint_fast32_t j = 0; j != runs_y; ++j){
      current_y += blocks.height(j);
      treat_block_row(row_table, col_table, j, current_y, blocks);
    }

    return row_table[nx];
  }

}
