
#pragma once

#include "utils.hpp"

namespace nn{

  uint_fast32_t nx, ny, runs_x, runs_y;

  template<class T>
  inline void treat_row(std::vector<T>& row, const uint_fast32_t j, const BlockInfo<T>& blocks)
  {
    uint_fast32_t index = 1;
    T last_cost = row[0] + blocks.cost(0,j);;
    for(uint_fast32_t i = 0; i != runs_x; ++i){
      const T cost = blocks.cost(i,j);
      for(uint_fast32_t x = 0; x != blocks.width(i); ++x){
        const T tmp = std::min(std::min(row[index-1], row[index]), last_cost) + cost;
        row[index - 1] = last_cost;
        last_cost = tmp;
        ++index;
      }
    }
    row[index - 1] = last_cost;
  }


  template<class T = double>
  T compute_dtw(const RLEncoding<T>& _x_seq, const RLEncoding<T>& _y_seq)
  {
    runs_x = _x_seq.size();
    runs_y = _y_seq.size();
    const BlockInfo<T> blocks(_x_seq, _y_seq);
    std::vector<T> row;
    
    for(uint_fast32_t j = 0; j < runs_y; ++j) ny += blocks.height(j);
    for(uint_fast32_t i = 0; i < runs_x; ++i) nx += blocks.width(i);
    row.reserve(nx + 1);

    T last_cost = 0;
    row.emplace_back(0);
    for(uint_fast32_t i = 0; i != runs_x; ++i){
      const T cost = blocks.cost(i,0);
      for(uint_fast32_t x = 0; x != blocks.width(i); ++x)
        row.emplace_back(last_cost += cost);
    }
    for(uint_fast32_t j = 0; j < runs_y; ++j)
      for(uint_fast32_t y = 0; y != blocks.height(j); ++y)
        treat_row(row,j,blocks);

    return row.back();
  }

}
