#pragma once

#include <iostream>
#include <sstream>
#include <string>
#include <utility>
#include <vector>
#include <iterator>

namespace akt {
    // Reads in the newline-separated non-empty time series in the UCR format
    // Reads at most maxSequencesToRead sequences, or all the sequences in the file if maxSequencesToRead < 0
    template <typename T = double>
    std::vector<std::vector<T>> parseUCR(std::istream& is, int maxSequencesToRead = -1)
    {
        std::vector<std::vector<T>> res;
        if (maxSequencesToRead == 0)
            return res;

        int sequencesRead = 0;
        for(std::string line; ((maxSequencesToRead < 0) || (sequencesRead < maxSequencesToRead)) && std::getline(is, line); ) {
            auto iss = std::istringstream(line);
            std::vector<T> sequence;
            std::copy(std::istream_iterator<T>(iss), std::istream_iterator<T>(), std::back_inserter(sequence));
            if(sequence.size() > 1) {
                // Ignore first element of the sequence since it is the class label (or a token for lack thereof)
                res.emplace_back(sequence.cbegin() + 1, sequence.cend());
                ++sequencesRead;
            }
        }

        return res;
    }
}
