
#pragma once

#include <iostream>
#include <list>
#include <vector>
#include<cassert>

#ifdef debuglevel
#define DEBUG(x) x
#else
#define DEBUG(x)
#endif

#ifdef STATISTICS
#define STAT(x) x
#else
#define STAT(x)
#endif

// translate (x,y) to linear coordinates for blocks
//#define idx(x,y) (runs_x * y + x)

using XY_Pos = std::pair<int_fast32_t, int_fast32_t>;

template<class T = double>
using RunlengthAndValue = std::pair<uint_fast32_t, T>;

template<class T = double>
using RLEncoding = std::vector<RunlengthAndValue<T>>;


template<class T = double>
std::ostream& operator<<(std::ostream& os, const RLEncoding<T>& x)
{
  for(const auto& p: x) os << "("<<p.first<<" x "<<p.second<<") ";
  return os;
}


// cells can be on a vertical or horizontal intersection (or both)
enum cell_class { vertical = 0x01, horizontal = 0x02, diagonal = 0x04};
// directions are the same as cell classes
using direction = cell_class;


// convert a time series to run-length encoding
template<class T = double>
RLEncoding<T> time_seq_to_RLE(const std::vector<T>& time_seq)
{
  RLEncoding<T> res;
  res.emplace_back(0, time_seq[0]);
  for(const T& val: time_seq)
    if(val == res.back().second){
      ++res.back().first;
    } else res.emplace_back(1, val);
  return res;
}


// return information about blocks (width/height/cost)
template<class T = double>
struct BlockInfo
{
  const RLEncoding<T>& x_seq;
  const RLEncoding<T>& y_seq;

  BlockInfo(const RLEncoding<T>& _x, const RLEncoding<T>& _y):
    x_seq(_x), y_seq(_y)
  {}

  inline T cost(const uint_fast32_t x, const uint_fast32_t y) const
  {
    const T diff = x_seq[x].second - y_seq[y].second;
    return diff * diff;
  }
  inline uint_fast32_t height(const uint_fast32_t y) const { return y_seq[y].first; }
  inline uint_fast32_t width(const uint_fast32_t x) const { return x_seq[x].first; }
};

namespace std{
// output things
template<class T>
std::ostream& operator<<(std::ostream& os, const std::list<T>& l)
{
  for(const auto& x: l) os << x <<' ';
  return os;
}
template<class T, class Q>
std::ostream& operator<<(std::ostream& os, const std::pair<T,Q>& l)
{
  return os << "("<<l.first<<", "<<l.second<<")";
}
template<class T>
std::ostream& operator<<(std::ostream& os, const std::vector<T>& l)
{
  for(const auto& x: l) os << x <<' ';
  return os;
}
}


